@@ -23,6 +23,9 @@ import sys
 import unittest
 import warnings
 import getpass
+from collections.abc import MutableMapping
+
+PY310 = (3, 10) <= sys.version_info
 
 # Below is general helper stuff that Python uses in `unittests`.  As these
 # not meant for users, and could change without notice, include them
@@ -34,6 +37,87 @@ TESTFN = '@test'
 TESTFN = "{}_{}_tmp".format(TESTFN, os.getpid())
 
 
+class EnvironmentVarGuard(MutableMapping):
+    """
+    Class to help protect the environment variable properly.  Can be used as a context manager.
+
+    Directly ripped from Python support tools so that we can actually try and test `~user` expansion.
+
+    Taken from: https://github.com/python/cpython/blob/main/Lib/test/support/os_helper.py.
+    """
+
+    def __init__(self):
+        """Initialize."""
+
+        self._environ = os.environ
+        self._changed = {}
+
+    def __getitem__(self, envvar):
+        """Get item."""
+
+        return self._environ[envvar]
+
+    def __setitem__(self, envvar, value):
+        """Set item."""
+
+        # Remember the initial value on the first access
+        if envvar not in self._changed:
+            self._changed[envvar] = self._environ.get(envvar)
+        self._environ[envvar] = value
+
+    def __delitem__(self, envvar):
+        """Delete item."""
+
+        # Remember the initial value on the first access
+        if envvar not in self._changed:
+            self._changed[envvar] = self._environ.get(envvar)
+        if envvar in self._environ:
+            del self._environ[envvar]
+
+    def keys(self):
+        """Get keys."""
+        return self._environ.keys()
+
+    def __iter__(self):
+        """Iterate."""
+
+        return iter(self._environ)
+
+    def __len__(self):
+        """Get length."""
+        return len(self._environ)
+
+    def set(self, envvar, value):  # noqa: A003
+        """Set variable."""
+
+        self[envvar] = value
+
+    def unset(self, envvar):
+        """Unset variable."""
+
+        del self[envvar]
+
+    def copy(self):
+        """Copy environment."""
+
+        # We do what `os.environ.copy()` does.
+        return dict(self)
+
+    def __enter__(self):
+        """Enter."""
+        return self
+
+    def __exit__(self, *ignore_exc):
+        """Exit."""
+        for (k, v) in self._changed.items():
+            if v is None:
+                if k in self._environ:
+                    del self._environ[k]
+            else:
+                self._environ[k] = v
+        os.environ = self._environ
+
+
 @contextlib.contextmanager
 def change_cwd(path, quiet=False):
     """
@@ -916,6 +1000,18 @@ class Testglob(_TestGlob):
         for file in glob.glob(b'!**/', flags=glob.N | glob.NEGATEALL | glob.G, root_dir=os.fsencode(self.tempdir)):
             self.assert_equal(os.path.isdir(file), False)
 
+    def test_magic_non_magic(self):
+        """Test logic when switching from magic to non-magic patterns."""
+
+        with change_cwd(self.tempdir):
+            self.assert_equal(sorted(glob.glob(['**/aab', 'dummy'], flags=glob.G)), ['aab',])
+
+    def test_non_magic_magic(self):
+        """Test logic when switching from non-magic to magic patterns."""
+
+        with change_cwd(self.tempdir):
+            self.assert_equal(sorted(glob.glob(['dummy', '**/aab'], flags=glob.G)), ['aab',])
+
 
 class TestGlobMarked(Testglob):
     """Test glob marked."""
@@ -1507,13 +1603,29 @@ class TestSymlinkLoopGlob(unittest.TestCase):
 class TestGlobPaths(unittest.TestCase):
     """Test `glob` paths."""
 
-    def test_root(self):
+    @unittest.skipUnless(not sys.platform.startswith('win'), "Linux/Unix specific test")
+    def test_root_unix(self):
+        """Test that `glob` translates the root properly."""
+
+        # On Linux/Unix, this should translate to the root.
+        # Basically, we should not return an empty set.
+        results = glob.glob('/*')
+        self.assertTrue(len(results) > 0)
+        self.assertTrue('/' not in results)
+
+    @unittest.skipUnless(sys.platform.startswith('win'), "Windows specific test")
+    def test_root_win(self):
         """Test that `glob` translates the root properly."""
 
         # On Windows, this should translate to the current drive.
-        # On Linux/Unix, this should translate to the root.
         # Basically, we should not return an empty set.
-        self.assertTrue(len(glob.glob('/*')) > 0)
+        results = glob.glob('/*')
+        self.assertTrue(len(results) > 0)
+        self.assertTrue('\\' not in results)
+
+        results = glob.glob(r'\\*')
+        self.assertTrue(len(results) > 0)
+        self.assertTrue('\\' not in results)
 
     def test_start(self):
         """Test that starting directory/files are handled properly."""
@@ -1524,6 +1636,16 @@ class TestGlobPaths(unittest.TestCase):
         )
 
 
+class TestExcludes(unittest.TestCase):
+    """Test expansion limits."""
+
+    def test_translate_exclude(self):
+        """Test excludes."""
+
+        results = glob.glob('**/*.md', exclude='**/README.md', flags=glob.GLOBSTAR)
+        self.assertTrue('README.md' not in results)
+
+
 @unittest.skipUnless(os.path.expanduser('~') != '~', "Requires expand user functionality")
 class TestTilde(unittest.TestCase):
     """Test tilde cases."""
@@ -1543,13 +1665,34 @@ class TestTilde(unittest.TestCase):
     def test_tilde_user(self):
         """Test tilde user cases."""
 
-        try:
-            user = getpass.getuser()
-        except ModuleNotFoundError:
-            user = os.path.basename(os.path.expanduser('~'))
+        if sys.platform.startswith('win') and PY310:
+            # In CI, and maybe on other systems, we cannot be sure we'll be able to get the user.
+            # So fake it by using our own user name with the current `USERPROFILE` path.
+            with EnvironmentVarGuard() as env:
+                userpath = os.environ.get('USERPROFILE')
+                env.clear()
+                user = 'test_user'
+                env['USERPROFILE'] = userpath
+                env['USERNAME'] = 'test_user'
 
-        files = os.listdir(os.path.expanduser('~{}'.format(user)))
-        self.assertEqual(len(glob.glob('~{}/*'.format(user), flags=glob.T | glob.D)), len(files))
+                files = os.listdir(userpath)
+                self.assertEqual(len(glob.glob('~{}/*'.format(user), flags=glob.T | glob.D)), len(files))
+        else:
+            # Accommodate non-Windows user behavior
+            user = None
+            if not sys.platform.startswith('win'):
+                try:
+                    user = getpass.getuser()
+                except ModuleNotFoundError:
+                    pass
+
+            if user is None:
+                # Last ditch effort to get a user.
+                user = os.path.basename(os.path.expanduser('~'))
+
+            userpath = os.path.expanduser('~{}'.format(user))
+            files = os.listdir(userpath)
+            self.assertEqual(len(glob.glob('~{}/*'.format(user), flags=glob.T | glob.D)), len(files))
 
     def test_tilde_disabled(self):
         """Test when tilde is disabled."""
@@ -1623,3 +1766,8 @@ class TestInputTypes(unittest.TestCase):
 
         with self.assertRaises(TypeError):
             glob.glob('docs/*', root_dir=b'.')
+
+    def test_cwd_root_dir_empty(self):
+        """Test empty patterns with current working directory."""
+
+        self.assertEqual(glob.glob([], root_dir='.'), [])
