@@ -2,28 +2,15 @@
 Wild Card Match.
 
 A module for performing wild card matches.
-
-Licensed under MIT
-Copyright (c) 2018 - 2020 Isaac Muse <isaacmuse@gmail.com>
-
-Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
-documentation files (the "Software"), to deal in the Software without restriction, including without limitation
-the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
-and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
-
-The above copyright notice and this permission notice shall be included in all copies or substantial portions
-of the Software.
-
-THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
-TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
-THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
-CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
-IN THE SOFTWARE.
 """
+from __future__ import annotations
 import os
 import re
 from . import _wcparse
+from . import _wcmatch
 from . import util
+from typing import Any, Iterator, Generic, AnyStr
+
 
 __all__ = (
     "CASE", "IGNORECASE", "RAWCHARS", "FILEPATHNAME", "DIRPATHNAME", "PATHNAME",
@@ -78,37 +65,62 @@ FLAG_MASK = (
 )
 
 
-class WcMatch:
+class WcMatch(Generic[AnyStr]):
     """Finds files by wildcard."""
 
-    def __init__(self, root_dir, file_pattern=None, exclude_pattern=None, flags=0, limit=_wcparse.PATHNAME, **kwargs):
+    def __init__(
+        self,
+        root_dir: AnyStr,
+        file_pattern: AnyStr | None = None,
+        exclude_pattern: AnyStr | None = None,
+        flags: int = 0,
+        limit: int = _wcparse.PATHNAME,
+        **kwargs: Any
+    ):
         """Initialize the directory walker object."""
 
+        self.is_bytes = isinstance(root_dir, bytes)
+        self._directory = self._norm_slash(root_dir)  # type: AnyStr
         self._abort = False
         self._skipped = 0
-        self._directory = _wcparse.norm_slash(root_dir, 0)
-        self.is_bytes = isinstance(self._directory, bytes)
-        if not self._directory:
-            if self.is_bytes:
-                curdir = bytes(os.curdir, 'ASCII')
-            else:
-                curdir = os.curdir
-        else:
-            curdir = self._directory
-        self.sep = os.fsencode(os.sep) if self.is_bytes else os.sep
-        self._root_dir = curdir if curdir.endswith(self.sep) else curdir + self.sep
-        self.file_pattern = file_pattern
-        if not self.file_pattern:
-            self.file_pattern = _wcparse.WcRegexp(
-                (re.compile(br'^.*$', re.DOTALL),) if self.is_bytes else (re.compile(r'^.*$', re.DOTALL),)
-            )
-        self.exclude_pattern = exclude_pattern if exclude_pattern is not None else (b'' if self.is_bytes else '')
         self._parse_flags(flags)
+        self._sep = os.fsencode(os.sep) if isinstance(root_dir, bytes) else os.sep  # type: AnyStr
+        self._root_dir = self._add_sep(self._get_cwd(), True)  # type: AnyStr
         self.limit = limit
+        empty = os.fsencode('') if isinstance(root_dir, bytes) else ''
+        self.pattern_file = file_pattern if file_pattern is not None else empty  # type: AnyStr
+        self.pattern_folder_exclude = exclude_pattern if exclude_pattern is not None else empty  # type: AnyStr
+        self.file_check = None  # type: _wcmatch.WcRegexp[AnyStr] | None
+        self.folder_exclude_check = None  # type: _wcmatch.WcRegexp[AnyStr] | None
         self.on_init(**kwargs)
-        self.file_check, self.folder_exclude_check = self._compile(self.file_pattern, self.exclude_pattern)
+        self._compile(self.pattern_file, self.pattern_folder_exclude)
 
-    def _parse_flags(self, flags):
+    def _norm_slash(self, name: AnyStr) -> AnyStr:
+        """Normalize path slashes."""
+
+        if util.is_case_sensitive():
+            return name
+        elif isinstance(name, bytes):
+            return name.replace(b'/', b"\\")
+        else:
+            return name.replace('/', "\\")
+
+    def _add_sep(self, path: AnyStr, check: bool = False) -> AnyStr:
+        """Add separator."""
+
+        return (path + self._sep) if not check or not path.endswith(self._sep) else path
+
+    def _get_cwd(self) -> AnyStr:
+        """Get current working directory."""
+
+        if self._directory:
+            return self._directory
+        elif isinstance(self._directory, bytes):
+            return bytes(os.curdir, 'ASCII')
+        else:
+            return os.curdir
+
+    def _parse_flags(self, flags: int) -> None:
         """Parse flags."""
 
         self.flags = flags & FLAG_MASK
@@ -123,7 +135,7 @@ class WcMatch:
             self.flags |= _FORCEWIN
         self.flags = self.flags & (_wcparse.FLAG_MASK ^ MATCHBASE)
 
-    def _compile_wildcard(self, pattern, pathname=False):
+    def _compile_wildcard(self, pattern: AnyStr, pathname: bool = False) -> _wcmatch.WcRegexp[AnyStr] | None:
         """Compile or format the wildcard inclusion/exclusion pattern."""
 
         flags = self.flags
@@ -132,20 +144,26 @@ class WcMatch:
             if self.matchbase:
                 flags |= MATCHBASE
 
-        return _wcparse.compile(pattern, flags, self.limit) if pattern else None
+        return _wcparse.compile([pattern], flags, self.limit) if pattern else None
 
-    def _compile(self, file_pattern, folder_exclude_pattern):
+    def _compile(self, file_pattern: AnyStr, folder_exclude_pattern: AnyStr) -> None:
         """Compile patterns."""
 
-        if not isinstance(file_pattern, _wcparse.WcRegexp):
-            file_pattern = self._compile_wildcard(file_pattern, self.file_pathname)
+        if self.file_check is None:
+            if not file_pattern:
+                self.file_check = _wcmatch.WcRegexp(
+                    (re.compile(br'^.*$' if isinstance(file_pattern, bytes) else r'^.*$', re.DOTALL),)
+                )
+            else:
+                self.file_check = self._compile_wildcard(file_pattern, self.file_pathname)
 
-        if not isinstance(folder_exclude_pattern, _wcparse.WcRegexp):
-            folder_exclude_pattern = self._compile_wildcard(folder_exclude_pattern, self.dir_pathname)
+        if self.folder_exclude_check is None:
+            if not folder_exclude_pattern:
+                self.folder_exclude_check = _wcmatch.WcRegexp(tuple())
+            else:
+                self.folder_exclude_check = self._compile_wildcard(folder_exclude_pattern, self.dir_pathname)
 
-        return file_pattern, folder_exclude_pattern
-
-    def _valid_file(self, base, name):
+    def _valid_file(self, base: AnyStr, name: AnyStr) -> bool:
         """Return whether a file can be searched."""
 
         valid = False
@@ -156,17 +174,17 @@ class WcMatch:
             valid = False
         return self.on_validate_file(base, name) if valid else valid
 
-    def compare_file(self, filename):
+    def compare_file(self, filename: AnyStr) -> bool:
         """Compare filename."""
 
-        return self.file_check.match(filename)
+        return self.file_check.match(filename)  # type: ignore[union-attr]
 
-    def on_validate_file(self, base, name):
+    def on_validate_file(self, base: AnyStr, name: AnyStr) -> bool:
         """Validate file override."""
 
         return True
 
-    def _valid_folder(self, base, name):
+    def _valid_folder(self, base: AnyStr, name: AnyStr) -> bool:
         """Return whether a folder can be searched."""
 
         valid = True
@@ -174,7 +192,7 @@ class WcMatch:
         if (
             not self.recursive or
             (
-                self.folder_exclude_check is not None and
+                self.folder_exclude_check and
                 not self.compare_directory(fullpath[self._base_len:] if self.dir_pathname else name)
             )
         ):
@@ -183,58 +201,60 @@ class WcMatch:
             valid = False
         return self.on_validate_directory(base, name) if valid else valid
 
-    def compare_directory(self, directory):
+    def compare_directory(self, directory: AnyStr) -> bool:
         """Compare folder."""
 
-        return not self.folder_exclude_check.match(directory + self.sep if self.dir_pathname else directory)
+        return not self.folder_exclude_check.match(  # type: ignore[union-attr]
+            self._add_sep(directory) if self.dir_pathname else directory
+        )
 
-    def on_init(self, **kwargs):
+    def on_init(self, **kwargs: Any) -> None:
         """Handle custom initialization."""
 
-    def on_validate_directory(self, base, name):
+    def on_validate_directory(self, base: AnyStr, name: AnyStr) -> bool:
         """Validate folder override."""
 
         return True
 
-    def on_skip(self, base, name):
+    def on_skip(self, base: AnyStr, name: AnyStr) -> Any:
         """On skip."""
 
         return None
 
-    def on_error(self, base, name):
+    def on_error(self, base: AnyStr, name: AnyStr) -> Any:
         """On error."""
 
         return None
 
-    def on_match(self, base, name):
+    def on_match(self, base: AnyStr, name: AnyStr) -> Any:
         """On match."""
 
         return os.path.join(base, name)
 
-    def on_reset(self):
+    def on_reset(self) -> None:
         """On reset."""
 
-    def get_skipped(self):
+    def get_skipped(self) -> int:
         """Get number of skipped files."""
 
         return self._skipped
 
-    def kill(self):
+    def kill(self) -> None:
         """Abort process."""
 
         self._abort = True
 
-    def is_aborted(self):
+    def is_aborted(self) -> bool:
         """Check if process has been aborted."""
 
         return self._abort
 
-    def reset(self):
+    def reset(self) -> None:
         """Revive class from a killed state."""
 
         self._abort = False
 
-    def _walk(self):
+    def _walk(self) -> Iterator[Any]:
         """Start search for valid files."""
 
         self._base_len = len(self._root_dir)
@@ -280,12 +300,12 @@ class WcMatch:
                     if self.is_aborted():
                         break
 
-    def match(self):
+    def match(self) -> list[Any]:
         """Run the directory walker."""
 
         return list(self.imatch())
 
-    def imatch(self):
+    def imatch(self) -> Iterator[Any]:
         """Run the directory walker as iterator."""
 
         self.on_reset()
